import { plejdChalResp as plejdCharResp, plejdEncodeDecode, reverseBuffer, } from "./plejdUtils.js";
import { randomBytes } from "crypto";
import noble from "@abandonware/noble";
import { PLEJD_PING_TIMEOUT, PLEJD_WRITE_TIMEOUT } from "./settings.js";
import { delay, race } from "./utils.js";
/**
 * Plejd BLE UUIDs
 */
var PlejdCharacteristics;
(function (PlejdCharacteristics) {
    PlejdCharacteristics["Service"] = "31ba000160854726be45040c957391b5";
    PlejdCharacteristics["LightLevel"] = "31ba000360854726be45040c957391b5";
    PlejdCharacteristics["Data"] = "31ba000460854726be45040c957391b5";
    PlejdCharacteristics["LastData"] = "31ba000560854726be45040c957391b5";
    PlejdCharacteristics["Auth"] = "31ba000960854726be45040c957391b5";
    PlejdCharacteristics["Ping"] = "31ba000a60854726be45040c957391b5";
})(PlejdCharacteristics || (PlejdCharacteristics = {}));
var PlejdCommand;
(function (PlejdCommand) {
    PlejdCommand["OnOffState"] = "0097";
    PlejdCommand["StateBrightness"] = "00c8";
    PlejdCommand["Brightness"] = "0098";
    PlejdCommand["Scene"] = "0021";
    PlejdCommand["Time"] = "001b";
    PlejdCommand["RequestResponse"] = "0102";
    PlejdCommand["RequestNoResponse"] = "0110";
    PlejdCommand["RequestReadValue"] = "0103";
    PlejdCommand["ButtonClick"] = "0006";
})(PlejdCommand || (PlejdCommand = {}));
export class PlejdService {
    config;
    log;
    onUpdate;
    sendQueue = [];
    plejdTimeout = null;
    queueTimeout = null;
    blacklistCleanupInterval = null;
    isConnecting = false;
    isAuthenticated = false;
    deviceState = {
        lastPingSuccess: Date.now(),
        consecutivePingFailures: 0,
    };
    deviceBlacklist = new Map();
    MAX_FAILURES = 5;
    BLACKLIST_DURATION = 5 * 60 * 1000;
    EXTENDED_BLACKLIST_DURATION = 30 * 60 * 1000;
    constructor(config, log, onUpdate) {
        this.config = config;
        this.log = log;
        this.onUpdate = onUpdate;
    }
    /**
     *
     * Update the state of a device
     *
     * @param identifier: The device identifier
     * @param isOn: The new state of the device
     * @param brightness: The new brightness of the device between 0-100
     */
    updateState = async (identifier, isOn, brightness) => {
        const payload = Buffer.from(!brightness || brightness === 0
            ? identifier.toString(16).padStart(2, "0") +
                PlejdCommand.RequestNoResponse +
                PlejdCommand.OnOffState +
                (isOn ? "01" : "00")
            : identifier.toString(16).padStart(2, "0") +
                PlejdCommand.RequestNoResponse +
                PlejdCommand.Brightness +
                "01" +
                Math.round(2.55 * brightness)
                    .toString(16)
                    .padStart(4, "0"), "hex");
        this.sendQueue.unshift(payload);
    };
    configureBLE = () => {
        this.startBlacklistCleanup();
        noble.on("stateChange", async (state) => {
            this.log.debug(`Noble State changed: ${state}`);
            await this.tryStartScanning();
        });
        noble.on("warning", (msg) => {
            this.log.warn("Noble warning: ", msg);
        });
    };
    //   -------------- Private -------------- \\
    onDiscover = async (peripheral) => {
        if (peripheral.rssi < -90) {
            this.log.debug(`Skipping device with weak signal: ${peripheral.address} (${peripheral.rssi} dB)`);
            return;
        }
        if (this.isDeviceBlacklisted(peripheral.address)) {
            this.log.debug(`Skipping blacklisted device: ${peripheral.address}`);
            return;
        }
        if (this.isConnecting) {
            this.log.debug("Connection already in progress, skipping...");
            return;
        }
        this.log.info(`Discovered | ${peripheral.advertisement.localName} | addr: ${peripheral.address} | Signal strength: ${this.mapRssiToQuality(peripheral.rssi)} (${peripheral.rssi} dB)`);
        this.log.debug(`Stopping scan`);
        await noble.stopScanningAsync();
        let retryCount = 0;
        const maxRetries = 3;
        const connectWithRetry = async () => {
            try {
                if (retryCount > 0) {
                    await delay(2000);
                }
                this.isConnecting = true;
                this.log.debug(`Connecting to the new peripheral`);
                await peripheral.connectAsync();
                this.log.info(`Connected to mesh | ${peripheral.advertisement.localName} (addr: ${peripheral.address})`);
                peripheral.once("disconnect", async () => {
                    this.log.info("Disconnected from mesh");
                    this.cleanup();
                    if (retryCount < maxRetries) {
                        retryCount++;
                        this.log.debug(`Attempting to reconnect (attempt ${retryCount})`);
                        await connectWithRetry();
                    }
                    else {
                        this.log.error("Max reconnection attempts reached. Starting scan.");
                        await delay(5000);
                        await this.tryStartScanning();
                    }
                });
                // Reset retry count on successful connection
                retryCount = 0;
                let characteristics;
                try {
                    characteristics = await this.discoverCaracteristics(peripheral);
                }
                catch (e) {
                    this.log.error("Failed to discover characteristics, disconnecting. Error:", e);
                    await this.tryDisconnect(peripheral);
                    throw e;
                }
                await this.setupDevice(peripheral, characteristics);
            }
            catch (error) {
                this.log.error(`Connecting failed | ${peripheral.advertisement.localName} | addr: ${peripheral.address}) - err: ${error}`);
                await this.tryDisconnect(peripheral);
                if (retryCount < maxRetries) {
                    retryCount++;
                    this.log.debug(`Attempting to reconnect (attempt ${retryCount})`);
                    await connectWithRetry();
                }
                else {
                    this.log.error("Max reconnection attempts reached. Resetting noble and starting scan.");
                    noble.reset();
                    await this.tryStartScanning();
                }
            }
            finally {
                this.isConnecting = false;
            }
        };
        await connectWithRetry();
    };
    mapRssiToQuality(rssi) {
        if (rssi >= -30) {
            return "Excellent (Very close proximity)";
        }
        else if (rssi >= -67) {
            return "Very Good";
        }
        else if (rssi >= -70) {
            return "Good";
        }
        else if (rssi >= -80) {
            return "Fair";
        }
        else if (rssi >= -90) {
            return "Weak";
        }
        else {
            return "Very Weak (Potential connection issues)";
        }
    }
    discoverCaracteristics = async (peripheral) => {
        this.log.debug("Discovering characteristics");
        const services = [PlejdCharacteristics.Service];
        const characteristicIds = [
            PlejdCharacteristics.Data,
            PlejdCharacteristics.LastData,
            PlejdCharacteristics.Auth,
            PlejdCharacteristics.Ping,
        ];
        const { characteristics } = await peripheral.discoverSomeServicesAndCharacteristicsAsync(services, characteristicIds);
        this.log.debug("Characteristics discovered");
        return characteristics;
    };
    setupDevice = async (peripheral, characteristics) => {
        this.log.debug("Locating relevant characteristics");
        const authChar = characteristics.find((char) => char.uuid === PlejdCharacteristics.Auth);
        const lastDataChar = characteristics.find((char) => char.uuid === PlejdCharacteristics.LastData);
        const pingChar = characteristics.find((char) => char.uuid === PlejdCharacteristics.Ping);
        const dataChar = peripheral?.services[0]?.characteristics?.find((char) => char.uuid === PlejdCharacteristics.Data) ?? null;
        if (!authChar || !lastDataChar || !pingChar || !dataChar) {
            this.log.error("Unable to extract characteristic during discovery", authChar, lastDataChar, pingChar);
            await this.tryDisconnect(peripheral);
            return;
        }
        const addressBuffer = reverseBuffer(Buffer.from(String(peripheral.address).replace(/:/g, ""), "hex"));
        await this.authenticate(peripheral, authChar);
        await this.setupCommunication(peripheral, pingChar, lastDataChar, dataChar, addressBuffer);
    };
    handleQueuedMessages = (dataChar, addressBuffer) => {
        if (this.queueTimeout) {
            clearTimeout(this.queueTimeout);
        }
        this.queueTimeout = setTimeout(async () => {
            const payload = this.sendQueue.pop();
            if (!payload) {
                this.handleQueuedMessages(dataChar, addressBuffer);
                return;
            }
            const data = plejdEncodeDecode(this.config.cryptoKey, addressBuffer, payload);
            this.log.debug(`BLE command sent to ${addressBuffer?.toString("hex") ?? "Unknown"} | ${data.length} bytes | ${data.toString("hex")}`);
            let retryCount = 0;
            const maxRetries = 3;
            const tryWrite = async () => {
                try {
                    await race(() => dataChar.writeAsync(data, false));
                }
                catch (error) {
                    if (retryCount < maxRetries) {
                        retryCount++;
                        this.log.debug(`Retrying write (${retryCount}/${maxRetries})`);
                        await delay(100);
                        return tryWrite();
                    }
                    throw error;
                }
            };
            try {
                await tryWrite();
                this.handleQueuedMessages(dataChar, addressBuffer);
            }
            catch (error) {
                this.sendQueue.unshift(data);
                this.log.error("Failed to send data to device, will retry: ", error);
            }
        }, PLEJD_WRITE_TIMEOUT);
    };
    startPlejdPing = (peripheral, pingChar) => {
        if (this.plejdTimeout) {
            clearTimeout(this.plejdTimeout);
        }
        this.plejdTimeout = setTimeout(async () => {
            if (pingChar) {
                try {
                    const ping = randomBytes(1);
                    await race(() => pingChar.writeAsync(ping, false));
                    const pong = await race(() => pingChar.readAsync());
                    if (((ping[0] + 1) & 0xff) !== pong[0]) {
                        this.log.error("Ping pong communication failed, missing pong response");
                    }
                    this.deviceState.lastPingSuccess = Date.now();
                    this.deviceState.consecutivePingFailures = 0;
                    this.startPlejdPing(peripheral, pingChar);
                }
                catch (error) {
                    this.deviceState.consecutivePingFailures++;
                    const address = peripheral?.address;
                    if (address && this.deviceState.consecutivePingFailures >= 3) {
                        this.blacklistDevice(address, `Ping failed: ${error}`);
                        await this.tryDisconnect(peripheral);
                    }
                    this.log.warn("Ping failed, device disconnected, will retry to connect to mesh: ", error);
                }
            }
        }, PLEJD_PING_TIMEOUT);
    };
    handleNotification = async (data, isNotification, addressBuffer) => {
        const decodedData = plejdEncodeDecode(this.config.cryptoKey, addressBuffer, data);
        const id = parseInt(decodedData[0].toString(), 10);
        const command = decodedData.toString("hex", 3, 5);
        const isOn = parseInt(decodedData.toString("hex", 5, 6), 10) === 1;
        const commandType = Object.values(PlejdCommand).find((x) => x.toString() === command) ??
            "Unknown";
        const d = {
            id: id,
            command: command,
            commandType: commandType,
            on: isOn,
            Notification: isNotification,
            payload: decodedData.toString("hex"),
        };
        this.log.debug("Handle BLE notification", d);
        switch (command) {
            case PlejdCommand.Time: {
                const arg = parseInt(reverseBuffer(decodedData.subarray(5, 9)).toString("hex"), 16);
                const date = new Date(arg * 1000);
                this.log.debug("Time sync: " + date.toString());
                break;
            }
            case PlejdCommand.Brightness:
            case PlejdCommand.StateBrightness: {
                const dim = parseInt(decodedData.toString("hex", 7, 8), 16);
                // Convert to Homebridge 1-100
                const converted = dim === 0 ? 1 : (100 / 255) * dim;
                this.onUpdate(id, isOn, converted);
                break;
            }
            case PlejdCommand.Scene:
            case PlejdCommand.OnOffState:
            case PlejdCommand.ButtonClick:
            case PlejdCommand.RequestResponse:
            case PlejdCommand.RequestNoResponse:
            case PlejdCommand.RequestReadValue: {
                this.onUpdate(id, isOn);
                break;
            }
            default: {
                this.onUpdate(id, isOn);
                this.log.warn(`Unknown | command: ${command} | id: ${id} | ${decodedData.toString("hex")}`);
            }
        }
    };
    setupCommunication = async (peripheral, pingChar, lastDataChar, dataChar, addressBuffer) => {
        this.log.debug("Setting up ping pong communication with Plejd device");
        this.startPlejdPing(peripheral, pingChar);
        this.log.debug("Starting queue handler for messages to Plejd device");
        this.handleQueuedMessages(dataChar, addressBuffer);
        try {
            this.log.debug("Subscribing to incomming messages");
            await race(() => lastDataChar.subscribeAsync());
            lastDataChar.on("data", async (data, isNotification) => {
                await this.handleNotification(data, isNotification, addressBuffer);
            });
        }
        catch (e) {
            const address = peripheral?.address;
            if (address) {
                this.blacklistDevice(address, `Communication setup failed: ${e}`);
            }
            await this.tryDisconnect(peripheral);
            this.log.error("Failed to subscribe to Plejd device", e);
        }
    };
    authenticate = async (peripheral, authChar) => {
        if (this.isAuthenticated) {
            return;
        }
        try {
            this.log.debug("Authenticating to Plejd device");
            await race(() => authChar.writeAsync(Buffer.from([0x00]), false));
            await delay(100);
            const data = await race(() => authChar.readAsync());
            await delay(100);
            await authChar.writeAsync(plejdCharResp(this.config.cryptoKey, data), false);
            await delay(100);
            this.log.debug("Authentication successful");
            this.isAuthenticated = true;
        }
        catch (e) {
            this.isAuthenticated = false;
            const address = peripheral?.address;
            if (address) {
                this.blacklistDevice(address, `Authentication failed: ${e}`);
            }
            await this.tryDisconnect(peripheral);
            this.log.error("Failed to authenticate to Plejd device", e);
        }
    };
    tryStartScanning = async () => {
        try {
            if (noble._state === "poweredOn") {
                this.cleanupBlacklist();
                this.log.info("Scanning for Plejd devices");
                await delay(5000);
                await race(() => noble.startScanningAsync([PlejdCharacteristics.Service], false), 10_000);
                noble.once("discover", async (peripheral) => await this.onDiscover(peripheral));
                setTimeout(async () => {
                    if (!this.isAuthenticated) {
                        this.log.warn("No device found during scan, restarting scan");
                        await noble.stopScanningAsync();
                        await this.tryStartScanning();
                    }
                }, 30000);
            }
        }
        catch (e) {
            this.log.error("Failed to start scanning for Plejd devices", e);
            await delay(10000);
            await this.tryStartScanning();
        }
    };
    tryDisconnect = async (peripheral) => {
        try {
            await peripheral.disconnectAsync();
        }
        catch (e) {
            this.log.debug("Failed to disconnect from previous peripheral", e);
        }
    };
    isDeviceBlacklisted(address) {
        const blacklistEntry = this.deviceBlacklist.get(address);
        if (!blacklistEntry)
            return false;
        if (Date.now() > blacklistEntry.until) {
            this.deviceBlacklist.delete(address);
            return false;
        }
        return true;
    }
    blacklistDevice(address, reason) {
        const entry = this.deviceBlacklist.get(address) ?? {
            until: 0,
            attempts: 0,
        };
        entry.attempts++;
        const baseTime = entry.attempts >= this.MAX_FAILURES
            ? this.EXTENDED_BLACKLIST_DURATION
            : this.BLACKLIST_DURATION;
        const duration = Math.min(baseTime * Math.pow(2, entry.attempts - 1), 24 * 60 * 60 * 1000);
        entry.until = Date.now() + duration;
        this.deviceBlacklist.set(address, entry);
        this.log.warn(`Device ${address} blacklisted for ${duration / 1000} seconds. Reason: ${reason}. Failure attempts: ${entry.attempts}`);
    }
    cleanupBlacklist() {
        const now = Date.now();
        for (const [address, entry] of this.deviceBlacklist.entries()) {
            if (now > entry.until) {
                this.deviceBlacklist.delete(address);
            }
        }
    }
    startBlacklistCleanup() {
        if (this.blacklistCleanupInterval) {
            clearInterval(this.blacklistCleanupInterval);
        }
        this.blacklistCleanupInterval = setInterval(() => {
            this.cleanupBlacklist();
        }, 5 * 60 * 1000); // Clean every 5 minutes
    }
    cleanup() {
        if (this.plejdTimeout) {
            clearTimeout(this.plejdTimeout);
            this.plejdTimeout = null;
        }
        if (this.queueTimeout) {
            clearTimeout(this.queueTimeout);
            this.queueTimeout = null;
        }
        if (this.blacklistCleanupInterval) {
            clearInterval(this.blacklistCleanupInterval);
            this.blacklistCleanupInterval = null;
        }
        this.sendQueue = [];
        this.isAuthenticated = false;
        this.deviceState = {
            lastPingSuccess: Date.now(),
            consecutivePingFailures: 0,
        };
    }
}
//# sourceMappingURL=plejdService.js.map